#include <torch/extension.h>
#include <ATen/ATen.h>
#include <cuda_runtime.h>

// Maximum number of floats allowed in constant memory (16384 floats ~ 64KB)
#define MAX_CLUSTERS_SIZE 16384

// Declare constant memory for the clusters tensor (assumed shape [D, K+G])
__constant__ float d_clusters[MAX_CLUSTERS_SIZE];

// Custom CUDA kernel that performs matrix multiplication using warp-level primitives
// Each warp computes one output element: dot(x[row,:], d_clusters[:, col]).
// x has dimensions [M, D] and clusters (in constant memory) is [D, N_out].
// The output out is of shape [M, N_out].
__global__ void matmul_warp_kernel(const float* __restrict__ x, float* __restrict__ out, int M, int D, int N_out) {
    // Calculate global warp id and the lane id within the warp
    int warp_id = (blockIdx.x * blockDim.x + threadIdx.x) / warpSize;
    int lane = threadIdx.x & 31;
    
    // Each warp is assigned to compute one element of the output matrix
    if (warp_id < M * N_out) {
        int row = warp_id / N_out;
        int col = warp_id % N_out;
        float sum = 0.0f;

        // Each lane accumulates a partial dot product over the D dimension
        // Stride via warpSize to cover the full vector
        for (int k = lane; k < D; k += warpSize) {
            // d_clusters is laid out in column-major order: index = k * N_out + col
            sum += x[row * D + k] * d_clusters[k * N_out + col];
        }

        // Use warp-level reduction with __shfl_down_sync to sum partial results
        for (int offset = warpSize / 2; offset > 0; offset /= 2) {
            sum += __shfl_down_sync(0xFFFFFFFF, sum, offset);
        }

        // Lane 0 writes the final result for this output element
        if (lane == 0) {
            out[row * N_out + col] = sum;
        }
    }
}

// Forward function for NetVLAD with ghost clusters using warp-level optimized matrix multiplication
// This function uses constant memory for clusters and a warp-level reduction for the x * clusters operation.
torch::Tensor forward(
    torch::Tensor x,
    torch::Tensor clusters,
    torch::Tensor clusters2,
    torch::Tensor bn_weight,
    torch::Tensor bn_bias,
    torch::Tensor bn_mean,
    torch::Tensor bn_var,
    bool is_training,
    int64_t cluster_size,
    int64_t feature_size
) {
    // x is expected to be [B, N, D]
    auto B = x.size(0);
    auto N = x.size(1);
    auto D = x.size(2);
    TORCH_CHECK(D == feature_size, "feature_size mismatch.");

    // Flatten x to [M, D] where M = B * N
    int64_t M = B * N;
    auto x_flat = x.reshape({M, D}).contiguous();

    // clusters is of shape [D, (K+G)] with (K+G) >= cluster_size
    int N_out = clusters.size(1);
    auto assignment = torch::empty({M, N_out}, x.options());

    // If clusters fits within constant memory, copy it there and use the warp-level kernel
    if (clusters.numel() <= MAX_CLUSTERS_SIZE) {
        auto clusters_contig = clusters.contiguous();
        size_t clusters_bytes = clusters_contig.numel() * sizeof(float);
        cudaError_t err = cudaMemcpyToSymbol(d_clusters, clusters_contig.data_ptr<float>(), clusters_bytes);
        TORCH_CHECK(err == cudaSuccess, "cudaMemcpyToSymbol failed: ", cudaGetErrorString(err));

        // Total output elements = M * N_out, each computed by one warp
        int total_warps = M * N_out;
        // Use 256 threads per block (i.e., 8 warps per block)
        int threadsPerBlock = 256;
        int warpsPerBlock = threadsPerBlock / 32;
        int blocks = (total_warps + warpsPerBlock - 1) / warpsPerBlock;

        matmul_warp_kernel<<<blocks, threadsPerBlock>>>(
            x_flat.data_ptr<float>(),
            assignment.data_ptr<float>(),
            M, D, N_out
        );
        err = cudaGetLastError();
        TORCH_CHECK(err == cudaSuccess, "Kernel launch failed: ", cudaGetErrorString(err));
        err = cudaDeviceSynchronize();
        TORCH_CHECK(err == cudaSuccess, "Kernel sync failed: ", cudaGetErrorString(err));
    } else {
        // Fallback to ATen's matmul if clusters do not fit in constant memory
        assignment = at::matmul(x_flat, clusters);
    }

    // Batch normalization
    assignment = at::batch_norm(
        assignment, bn_weight, bn_bias, bn_mean, bn_var,
        is_training, 0.1, 1e-5, true
    );

    // Softmax on dimension 1 and narrow out ghost clusters to obtain [M, cluster_size]
    assignment = at::softmax(assignment, 1);
    assignment = assignment.narrow(1, 0, cluster_size);

    // Reshape assignment to [B, N, cluster_size]
    assignment = assignment.reshape({B, N, cluster_size});

    // Compute a_sum as the sum of assignment along dim=1 -> [B, 1, cluster_size]
    auto a_sum = assignment.sum(1, true);
    // Multiply with clusters2 (expected shape: [1, D, cluster_size])
    auto a = a_sum * clusters2;

    // Transpose assignment to [B, cluster_size, N]
    assignment = assignment.transpose(1, 2);
    // Reshape x back to [B, N, D]
    auto x_reshaped = x_flat.reshape({B, N, D});

    // VLAD aggregation: perform batched matrix multiplication to get [B, cluster_size, D] then transpose to [B, D, cluster_size]
    auto vlad = at::bmm(assignment, x_reshaped);
    vlad = vlad.transpose(1, 2);
    vlad = vlad - a;

    // Intra-normalize across the D dimension
    vlad = vlad / (vlad.norm(2, {1}, true) + 1e-12);

    // Flatten to [B, D * cluster_size] and L2 normalize
    vlad = vlad.reshape({B, D * cluster_size});
    vlad = vlad / (vlad.norm(2, {1}, true) + 1e-12);

    return vlad;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &forward, "NetVLAD with ghost clusters (CUDA warp-level optimization)");
}
